"""Run pytest in scikit-bio configuration.

----------------------------------------------------------------------------
Copyright (c) 2013--, scikit-bio development team.

Distributed under the terms of the Modified BSD License.

The full license is in the file LICENSE.txt, distributed with this software.
----------------------------------------------------------------------------
"""

from skbio.util import pytestrunner


if __name__ == "__main__":
    pytestrunner()
