/*
----------------------------------------------------------------------------
Copyright (c) 2013--, scikit-bio development team.

Distributed under the terms of the Modified BSD License.

The full license is in the file LICENSE.txt, distributed with this software.
----------------------------------------------------------------------------
*/

/*
Custom stylesheet for the Sphinx documentation with the PyData theme.
Instructions:
https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/styling.html
*/

/* hide external link icon */

html {
  --pst-icon-external-link: unset; 

  --pst-font-size-h1: 2rem;
  --pst-font-size-h2: 1.5rem;
  --pst-font-size-h3: 1.25rem;
  --pst-font-size-h4: 1.1rem;
  --pst-font-size-h5: 1.0rem;
  --pst-font-size-h6: 1.0rem;
}


/* remove inline code borders and background */

code {
  border: none !important;
  background-color: transparent !important;
}


/* brand color (monochromatic: shamrock green) */

/*
primary color (for links, default: teal)
secondary color (for links on hover, default: violet)
target color (for highlights, default: sunset)
inline code color (default: violet)

Color variables are not well documented; but some of them can be found here:
https://github.com/pydata/pydata-sphinx-theme/blob/50fa47727af8d23757ddecd375c7a31cd282d248/pydata_sphinx_theme/static/css/theme.css#L43-L48
*/

html[data-theme="light"] {
  --pst-color-primary: #239552;
  --pst-color-target: #F5FDC6;
  --pst-color-secondary: var(--pst-color-primary);
  --pst-color-inline-code: var(--pst-color-accent);
  --pst-color-inline-code-links: var(--pst-color-primary);
  --pst-color-table-row-hover-bg: var(--pst-color-target);
}

html[data-theme="dark"] {
  --pst-color-primary: #72C093;
  --pst-color-target: #2C1E7F;
  --pst-color-secondary: var(--pst-color-primary);
  --pst-color-inline-code: var(--pst-color-accent);
  --pst-color-inline-code-links: var(--pst-color-primary);
  --pst-color-table-row-hover-bg: var(--pst-color-target);
}


/* banner background color (default: violet) */

html[data-theme="light"] div.bd-header-announcement {
  background-color: var(--pst-color-target);
}

html[data-theme="dark"] div.bd-header-announcement {
  background-color: var(--pst-color-target);
}
